import logging.config
import os
import warnings

from bamt.config import config

log_file_path = config.get(
    "LOG", "log_conf_loc", fallback="log_conf_path is not defined"
)

if not os.path.isdir(os.path.join(os.path.expanduser("~"), "BAMT")):
    os.mkdir(os.path.join(os.path.expanduser("~"), "BAMT"))

try:
    logging.config.fileConfig(log_file_path)
except BaseException:
    log_file_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "logging.conf"
    )
    logging.config.fileConfig(log_file_path)
    warnings.warn(
        "Reading log path location from config file failed. Default location will be used instead."
    )

logger_builder = logging.getLogger("builder")
logger_network = logging.getLogger("network")
logger_preprocessor = logging.getLogger("preprocessor")
logger_nodes = logging.getLogger("nodes")
logger_display = logging.getLogger("display")

logging.captureWarnings(True)
logger_warnings = logging.getLogger("py.warnings")



from math import log10
from random import choice

import pandas as pd
from golem.core.dag.graph_utils import ordered_subnodes_hierarchy
from numpy import std, mean, log
from scipy.stats import norm
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import train_test_split

from .CompositeModel import CompositeModel
from .MLUtils import MlModels


def custom_crossover_all_model(
    graph_first: CompositeModel, graph_second: CompositeModel, max_depth
):
    num_cros = 100
    try:
        for _ in range(num_cros):
            selected_node1 = choice(graph_first.nodes)
            if selected_node1.nodes_from is None or selected_node1.nodes_from == []:
                continue

            selected_node2 = graph_second.get_nodes_by_name(str(selected_node1))[0]
            if selected_node2.nodes_from is None or selected_node2.nodes_from == []:
                continue

            model1 = selected_node1.content["parent_model"]
            model2 = selected_node2.content["parent_model"]

            selected_node1.content["parent_model"] = model2
            selected_node2.content["parent_model"] = model1

            break

    except Exception as ex:
        print(ex)
    return graph_first, graph_second


def custom_mutation_add_structure(graph: CompositeModel, **kwargs):
    num_mut = 100
    try:
        for _ in range(num_mut):
            rid = choice(range(len(graph.nodes)))
            random_node = graph.nodes[rid]
            other_random_node = graph.nodes[choice(range(len(graph.nodes)))]
            nodes_not_cycling = random_node.descriptive_id not in [
                n.descriptive_id for n in ordered_subnodes_hierarchy(other_random_node)
            ] and other_random_node.descriptive_id not in [
                n.descriptive_id for n in ordered_subnodes_hierarchy(random_node)
            ]
            if nodes_not_cycling:
                other_random_node.nodes_from.append(random_node)
                ml_models = MlModels()
                other_random_node.content[
                    "parent_model"
                ] = ml_models.get_model_by_children_type(other_random_node)
                break

    except Exception as ex:
        graph.log.warn(f"Incorrect connection: {ex}")
    return graph


def custom_mutation_delete_structure(graph: CompositeModel, **kwargs):
    num_mut = 100
    try:
        for _ in range(num_mut):
            rid = choice(range(len(graph.nodes)))
            random_node = graph.nodes[rid]
            other_random_node = graph.nodes[choice(range(len(graph.nodes)))]
            if (
                random_node.nodes_from is not None
                and other_random_node in random_node.nodes_from
            ):
                random_node.nodes_from.remove(other_random_node)
                if not random_node.nodes_from:
                    random_node.content["parent_model"] = None
                break
    except Exception as ex:
        print(ex)
    return graph


def custom_mutation_reverse_structure(graph: CompositeModel, **kwargs):
    num_mut = 100
    try:
        for _ in range(num_mut):
            rid = choice(range(len(graph.nodes)))
            random_node = graph.nodes[rid]
            other_random_node = graph.nodes[choice(range(len(graph.nodes)))]
            if (
                random_node.nodes_from is not None
                and other_random_node in random_node.nodes_from
            ):
                random_node.nodes_from.remove(other_random_node)
                if not random_node.nodes_from:
                    random_node.content["parent_model"] = None
                other_random_node.nodes_from.append(random_node)
                ml_models = MlModels()
                other_random_node.content[
                    "parent_model"
                ] = ml_models.get_model_by_children_type(other_random_node)
                break
    except Exception as ex:
        print(ex)
    return graph


def custom_mutation_add_model(graph: CompositeModel, **kwargs):
    try:
        all_nodes = graph.nodes
        nodes_with_parents = [
            node
            for node in all_nodes
            if (node.nodes_from != [] and node.nodes_from is not None)
        ]
        if not nodes_with_parents:
            return graph
        node = choice(nodes_with_parents)
        ml_models = MlModels()
        node.content["parent_model"] = ml_models.get_model_by_children_type(node)
    except Exception as ex:
        print(ex)
    return graph


def composite_metric(graph: CompositeModel, data: pd.DataFrame, percent=0.02):
    data_all = data
    data_train, data_test = train_test_split(data_all, train_size=0.8, random_state=42)
    score, len_data = 0, len(data_train)
    for node in graph.nodes:
        data_of_node_train = data_train[node.content["name"]]
        data_of_node_test = data_test[node.content["name"]]
        if node.nodes_from is None or node.nodes_from == []:
            if node.content["type"] == "cont":
                mu, sigma = mean(data_of_node_train), std(data_of_node_train)
                score += norm.logpdf(
                    data_of_node_test.values, loc=mu, scale=sigma
                ).sum()
            else:
                count = data_of_node_train.value_counts()
                frequency = log(count / len_data)
                index = frequency.index.tolist()
                for value in data_of_node_test:
                    if value in index:
                        score += frequency[value]
        else:
            model, columns, target, idx = (
                MlModels().dict_models[node.content["parent_model"]](),
                [n.content["name"] for n in node.nodes_from],
                data_of_node_train.to_numpy(),
                data_train.index.to_numpy(),
            )
            setattr(model, "max_iter", 100000)
            features = data_train[columns].to_numpy()
            if len(set(target)) == 1:
                continue
            fitted_model = model.fit(features, target)

            features = data_test[columns].to_numpy()
            target = data_of_node_test.to_numpy()
            if node.content["type"] == "cont":
                predict = fitted_model.predict(features)
                mse = mean_squared_error(target, predict, squared=False) + 0.0000001
                a = norm.logpdf(target, loc=predict, scale=mse)
                score += a.sum()
            else:
                predict_proba = fitted_model.predict_proba(features)
                idx = pd.array(list(range(len(target))))
                li = []

                for i in idx:
                    a = predict_proba[i]
                    try:
                        b = a[target[i]]
                    except BaseException:
                        b = 0.0000001
                    if b < 0.0000001:
                        b = 0.0000001
                    li.append(log(b))
                score += sum(li)

    edges_count = len(graph.get_edges())
    score -= (edges_count * percent) * log10(len_data) * edges_count

    return -score



from datetime import timedelta
from typing import Dict, Optional, List, Tuple

from golem.core.adapter import DirectAdapter
from golem.core.dag.verification_rules import has_no_cycle, has_no_self_cycled_nodes
from golem.core.log import Log
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.crossover import CrossoverTypesEnum
from golem.core.optimisers.genetic.operators.inheritance import GeneticSchemeTypesEnum
from golem.core.optimisers.genetic.operators.selection import SelectionTypesEnum
from golem.core.optimisers.objective import Objective, ObjectiveEvaluate
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams
from pandas import DataFrame

from bamt.builders.builders_base import BaseDefiner
from bamt.utils import EvoUtils as evo


class EvoDefiner(BaseDefiner):
    """
    Object that might take additional methods to decompose structure builder class
    """

    def __init__(
        self,
        data: DataFrame,
        descriptor: Dict[str, Dict[str, str]],
        regressor: Optional[object] = None,
    ):
        super().__init__(data, descriptor, regressor)


class EvoStructureBuilder(EvoDefiner):
    """
    This class uses an evolutionary algorithm based on GOLEM to generate a Directed Acyclic Graph (DAG) that represents
    the structure of a Bayesian Network.

    Attributes:
        data (DataFrame): Input data used to build the structure.
        descriptor (dict): Descriptor describing node types and signs.
        regressor (object): A regression model for continuous nodes.
        has_logit (bool): Indicates whether a logit link function should be used.
        use_mixture (bool): Indicates whether a mixture model should be used.
    """

    def __init__(
        self,
        data: DataFrame,
        descriptor: Dict[str, Dict[str, str]],
        regressor: Optional[object],
        has_logit: bool,
        use_mixture: bool,
    ):
        super(EvoStructureBuilder, self).__init__(
            data=data, descriptor=descriptor, regressor=regressor
        )
        self.data = data
        self.descriptor = descriptor
        self.has_logit = has_logit
        self.use_mixture = use_mixture
        self.regressor = regressor
        self.params = {
            "init_edges": None,
            "init_nodes": None,
            "remove_init_edges": True,
            "white_list": None,
            "bl_add": None,
        }
        self.default_n_jobs = -1
        self.default_pop_size = 15
        self.default_crossover_prob = 0.9
        self.default_mutation_prob = 0.8
        self.default_max_arity = 100
        self.default_max_depth = 100
        self.default_timeout = 180
        self.default_num_of_generations = 50
        self.default_early_stopping_iterations = 50
        self.logging_level = 50
        self.objective_metric = evo.K2_metric
        self.default_crossovers = [
            CrossoverTypesEnum.exchange_edges,
            CrossoverTypesEnum.exchange_parents_one,
            CrossoverTypesEnum.exchange_parents_both,
        ]
        self.default_mutations = [
            evo.custom_mutation_add,
            evo.custom_mutation_delete,
            evo.custom_mutation_reverse,
        ]
        self.default_selection = [SelectionTypesEnum.tournament]
        self.default_constraints = [
            has_no_self_cycled_nodes,
            has_no_cycle,
            evo.has_no_duplicates,
        ]
        self.verbose = True

    def build(
        self,
        data: DataFrame,
        classifier: Optional[object],
        regressor: Optional[object],
        **kwargs
    ):
        """
        Calls the search method to execute all the evolutionary computations.

        Args:
            data (DataFrame): The data from which to build the structure.
            classifier (Optional[object]): A classification model for discrete nodes.
            regressor (Optional[object]): A regression model for continuous nodes.
        """
        best_graph_edge_list = self.search(data, **kwargs)

        # Convert the best graph to the format used by the Bayesian Network
        self.skeleton["V"] = self.vertices
        self.skeleton["E"] = best_graph_edge_list

        self.get_family()
        self.overwrite_vertex(
            has_logit=self.has_logit,
            use_mixture=self.use_mixture,
            classifier=classifier,
            regressor=regressor,
        )

    def search(self, data: DataFrame, **kwargs) -> List[Tuple[str, str]]:
        """
        Executes all the evolutionary computations and returns the best graph's edge list.

        Args:
            data (DataFrame): The data from which to build the structure.

        Returns:
            best_graph_edge_list (List[Tuple[str, str]]): The edge list of the best graph found by the search.
        """
        # Get the list of node names
        nodes_types = data.columns.to_list()

        # Create the initial population
        initial = [
            evo.CustomGraphModel(
                nodes=kwargs.get(
                    "init_nodes",
                    [evo.CustomGraphNode(node_type) for node_type in nodes_types],
                )
            )
        ]

        # Define the requirements for the evolutionary algorithm
        requirements = GraphRequirements(
            max_arity=kwargs.get("max_arity", self.default_max_arity),
            max_depth=kwargs.get("max_depth", self.default_max_depth),
            num_of_generations=kwargs.get(
                "num_of_generations", self.default_num_of_generations
            ),
            timeout=timedelta(minutes=kwargs.get("timeout", self.default_timeout)),
            early_stopping_iterations=kwargs.get(
                "early_stopping_iterations", self.default_early_stopping_iterations
            ),
            n_jobs=kwargs.get("n_jobs", self.default_n_jobs),
        )

        # Set the parameters for the evolutionary algorithm
        optimizer_parameters = GPAlgorithmParameters(
            pop_size=kwargs.get("pop_size", self.default_pop_size),
            crossover_prob=kwargs.get("crossover_prob", self.default_crossover_prob),
            mutation_prob=kwargs.get("mutation_prob", self.default_mutation_prob),
            genetic_scheme_type=GeneticSchemeTypesEnum.steady_state,
            mutation_types=kwargs.get("custom_mutations", self.default_mutations),
            crossover_types=kwargs.get("custom_crossovers", self.default_crossovers),
            selection_types=kwargs.get("selection_type", self.default_selection),
        )

        # Set the adapter for the conversion between the graph and the data
        # structures used by the optimizer
        adapter = DirectAdapter(
            base_graph_class=evo.CustomGraphModel, base_node_class=evo.CustomGraphNode
        )

        # Set the constraints for the graph

        constraints = kwargs.get("custom_constraints", [])

        constraints.extend(self.default_constraints)

        if kwargs.get("blacklist", None) is not None:
            constraints.append(evo.has_no_blacklist_edges)
        if kwargs.get("whitelist", None) is not None:
            constraints.append(evo.has_only_whitelist_edges)

        graph_generation_params = GraphGenerationParams(
            adapter=adapter,
            rules_for_constraint=constraints,
            available_node_types=nodes_types,
        )

        # Define the objective function to optimize
        objective = Objective(
            {"custom": kwargs.get("custom_metric", self.objective_metric)}
        )

        # Initialize the optimizer
        optimizer = EvoGraphOptimizer(
            objective=objective,
            initial_graphs=initial,
            requirements=requirements,
            graph_generation_params=graph_generation_params,
            graph_optimizer_params=optimizer_parameters,
        )

        # Define the function to evaluate the objective function
        objective_eval = ObjectiveEvaluate(objective, data=data)

        if not kwargs.get("verbose", self.verbose):
            Log().reset_logging_level(logging_level=50)

        # Run the optimization
        optimized_graph = optimizer.optimise(objective_eval)[0]

        # Get the best graph
        best_graph_edge_list = optimized_graph.operator.get_edges()
        best_graph_edge_list = self._convert_to_strings(best_graph_edge_list)

        return best_graph_edge_list

    @staticmethod
    def _convert_to_strings(nested_list):
        return [tuple([str(item) for item in inner_list]) for inner_list in nested_list]



# currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
# parentdir = os.path.dirname(currentdir)
# sys.path.insert(0,parentdir)
import numpy as np
import pandas as pd


def loc_to_DataFrame(data: np.array):
    """Function to convert array to DataFrame
    Args:
        data (np.array): input array

    Returns:
        data (pd.DataFrame): with string columns for filtering
    """
    nodes_type = get_type_numpy(data)
    if data.T.ndim == 1:
        data = data.T
        nodes_type = {0: nodes_type[0]}
    dtype = {
        key: "int64" if value == "disc" else "float64"
        for key, value in nodes_type.items()
        if value in ["disc", "cont"]
    }
    df = pd.DataFrame(data).astype(dtype)
    df.columns = df.columns.map(str)
    return df


def get_type_numpy(data: np.array):
    """Function to define the type of the columns of array
       disc - discrete node
       cont - continuous
    Args:
        data (np.array): input array

    Returns:
        dict: output dictionary where 'key' - node name and 'value' - node type
    Notes:
    -- You may have problems with confusing rows and columns
    """
    arr = data.T

    column_type = {}
    for i, row in enumerate(arr):
        if row.ndim == 0 or row.T.ndim == 0:
            row_is_integer = np.issubdtype(row, np.integer) or row.is_integer()
            column_type[i] = "disc" if row_is_integer else "cont"
        else:
            all_row_is_integer = all(
                np.issubdtype(x, np.integer) or x.is_integer() for x in row
            )
            column_type[i] = "disc" if all_row_is_integer else "cont"
        if column_type[i] not in ["disc", "cont"]:
            print("get_type_numpy: Incorrect type of row")
            print(row)
    return column_type



from typing import Dict, List, Tuple, Type

import networkx as nx
from pandas import DataFrame

from bamt.log import logger_preprocessor
from bamt.nodes.base import BaseNode


def nodes_types(data: DataFrame) -> Dict[str, str]:
    """
    Function to define the type of the node
           disc - discrete node
           cont - continuous
        Args:
            data: input dataset

        Returns:
            dict: output dictionary where 'key' - node name and 'value' - node type
    """

    column_type = dict()
    for c in data.columns.to_list():
        disc = ["str", "O", "b", "categorical", "object", "bool"]
        disc_numerical = ["int32", "int64"]
        cont = ["float32", "float64"]
        if data[c].dtype.name in disc:
            column_type[c] = "disc"
        elif data[c].dtype.name in cont:
            column_type[c] = "cont"
        elif data[c].dtype.name in disc_numerical:
            column_type[c] = "disc_num"
        else:
            logger_preprocessor.error(f"Unsupported data type. Dtype: {data[c].dtypes}")

    return column_type


def nodes_signs(nodes_types: dict, data: DataFrame) -> Dict[str, str]:
    """Function to define sign of the node
       neg - if node has negative values
       pos - if node has only positive values

    Args:
        data (pd.DataFrame): input dataset
        nodes_types (dict): dict with nodes_types

    Returns:
        dict: output dictionary where 'key' - node name and 'value' - sign of data
    """
    if list(nodes_types.keys()) != data.columns.to_list():
        logger_preprocessor.error("Nodes_types dictionary is not full.")
        return
    columns_sign = dict()
    for c in data.columns.to_list():
        if nodes_types[c] == "cont":
            if (data[c] < 0).any():
                columns_sign[c] = "neg"
            else:
                columns_sign[c] = "pos"
    return columns_sign


def get_descriptor(data) -> Dict[str, Dict[str, str]]:
    return {"types": nodes_types(data), "signs": nodes_signs(nodes_types(data), data)}


def toporder(nodes: List[Type[BaseNode]], edges: List[Tuple]) -> List[List[str]]:
    """
    Function for topological sorting
    """
    G = nx.DiGraph()
    G.add_nodes_from([node.name for node in nodes])
    G.add_edges_from(edges)
    return list(nx.topological_sort(G))


class GraphAnalyzer(object):
    """
    Object to analyze DAG.
    """

    def __init__(self, bn):
        self.bn = bn

    def _isolate_structure(self, nodes):
        isolated_edges = []
        for edge in self.bn.edges:
            if edge[0] in nodes and edge[1] in nodes:
                isolated_edges.append(edge)
        return isolated_edges

    def markov_blanket(self, node_name: str):
        node = self.bn[node_name]

        parents = node.cont_parents + node.disc_parents
        children = node.children
        fremd_eltern = []

        for child in node.children:
            all_parents = self.bn[child].cont_parents + self.bn[child].disc_parents

            if all_parents == [node_name]:
                continue
            else:
                new = all_parents
            fremd_eltern.extend(new)

        nodes = parents + children + fremd_eltern + [node_name]

        edges = self._isolate_structure(nodes)
        return {"nodes": list(set(nodes)), "edges": edges}

    def _collect_height(self, node_name, height):
        nodes = []
        node = self.bn[node_name]
        if height <= 0:
            return []

        if height == 1:
            return node.disc_parents + node.cont_parents

        for parent in node.cont_parents + node.disc_parents:
            nodes.append(parent)
            nodes.extend(self._collect_height(parent, height=height - 1))
        return nodes

    def _collect_depth(self, node_name, depth):
        nodes = []
        node = self.bn[node_name]

        if depth <= 0:
            return []

        if depth == 1:
            return node.children

        for child in node.children:
            nodes.append(child)
            nodes.extend(self._collect_depth(child, depth=depth - 1))

        return nodes

    def find_family(self, *args):
        node_name, height, depth, with_nodes = args
        if not with_nodes:
            with_nodes = []
        else:
            with_nodes = list(with_nodes)
        nodes = (
            self._collect_depth(node_name, depth)
            + self._collect_height(node_name, height)
            + [node_name]
        )

        nodes = list(set(nodes + with_nodes))

        return {"nodes": nodes, "edges": self._isolate_structure(nodes + with_nodes)}



from typing import Optional

from sklearn import linear_model

from .logit_node import LogitNode


class CompositeDiscreteNode(LogitNode):
    """
    Class for composite discrete node.
    """

    def __init__(self, name, classifier: Optional[object] = None):
        super(CompositeDiscreteNode, self).__init__(name)
        if classifier is None:
            classifier = linear_model.LogisticRegression(
                multi_class="multinomial", solver="newton-cg", max_iter=100
            )
        self.classifier = classifier
        self.type = "CompositeDiscrete" + f" ({type(self.classifier).__name__})"



import configparser
from os import path, stat

config = configparser.ConfigParser()

CONFIGFILE = path.join(path.dirname(path.abspath(__file__)), "selbst.ini")

if path.isfile(CONFIGFILE) and stat(CONFIGFILE).st_size != 0:
    config.read(CONFIGFILE)
else:
    open(CONFIGFILE, "a").close()
    config["LOG"] = {
        "log_conf_loc": path.join(path.dirname(path.abspath(__file__)), "logging.conf")
    }
    with open(CONFIGFILE, "w") as configfile:
        config.write(configfile)



from .base import BaseNetwork


class DiscreteBN(BaseNetwork):
    """
    Bayesian Network with Discrete Types of Nodes
    """

    def __init__(self):
        super(DiscreteBN, self).__init__()
        self.type = "Discrete"
        self.scoring_function = ""
        self._allowed_dtypes = ["disc", "disc_num"]
        self.has_logit = None
        self.use_mixture = None



import itertools
import math
import random
from typing import Dict, Optional, List, Union

import numpy as np
from pandas import DataFrame
from sklearn import linear_model
from sklearn.base import clone
from sklearn.metrics import mean_squared_error as mse

from .base import BaseNode
from .schema import CondGaussParams


class ConditionalGaussianNode(BaseNode):
    """
    Main class for Conditional Gaussian Node
    """

    def __init__(self, name, regressor: Optional[object] = None):
        super(ConditionalGaussianNode, self).__init__(name)
        if regressor is None:
            regressor = linear_model.LinearRegression()
        self.regressor = regressor
        self.type = "ConditionalGaussian" + f" ({type(self.regressor).__name__})"

    def fit_parameters(self, data: DataFrame) -> Dict[str, Dict[str, CondGaussParams]]:
        """
        Train params for Conditional Gaussian Node.
        Return:
        {"hybcprob": {<combination of outputs from discrete parents> : CondGaussParams}}
        """
        hycprob = dict()
        values = []
        combinations = []
        for d_p in self.disc_parents:
            values.append(np.unique(data[d_p].values))
        for xs in itertools.product(*values):
            combinations.append(list(xs))
        for comb in combinations:
            mask = np.full(len(data), True)
            for col, val in zip(self.disc_parents, comb):
                mask = (mask) & (data[col] == val)
            new_data = data[mask]
            key_comb = [str(x) for x in comb]
            if new_data.shape[0] > 1:
                if self.cont_parents:
                    model = clone(self.regressor)
                    model.fit(
                        new_data[self.cont_parents].values, new_data[self.name].values
                    )
                    predicted_value = model.predict(new_data[self.cont_parents].values)
                    variance = mse(
                        new_data[self.name].values, predicted_value, squared=False
                    )
                    hycprob[str(key_comb)] = {
                        "variance": variance,
                        "mean": np.nan,
                        "regressor_obj": model,
                        "regressor": type(self.regressor).__name__,
                        "serialization": None,
                    }
                else:
                    mean_base = np.mean(new_data[self.name].values)
                    variance = np.var(new_data[self.name].values)
                    hycprob[str(key_comb)] = {
                        "variance": variance,
                        "mean": mean_base,
                        "regressor_obj": None,
                        "regressor": None,
                        "serialization": None,
                    }
            else:
                hycprob[str(key_comb)] = {
                    "variance": np.nan,
                    "regressor": None,
                    "regressor_obj": None,
                    "serialization": None,
                    "mean": np.nan,
                }
        return {"hybcprob": hycprob}

    def get_dist(self, node_info, pvals):
        dispvals = []
        lgpvals = []
        for pval in pvals:
            if isinstance(pval, str):
                dispvals.append(pval)
            else:
                lgpvals.append(pval)

        lgdistribution = node_info["hybcprob"][str(dispvals)]

        # JOBLIB

        if self.cont_parents:
            flag = False
            for el in lgpvals:
                if str(el) == "nan":
                    flag = True
                    break
            if flag:
                return np.nan, np.nan
            else:
                if lgdistribution["regressor"]:
                    model = lgdistribution["regressor_obj"]

                    cond_mean = model.predict(np.array(lgpvals).reshape(1, -1))[0]
                    variance = lgdistribution["variance"]
                    return cond_mean, variance
                else:
                    return np.nan, np.nan

        else:
            return lgdistribution["mean"], math.sqrt(lgdistribution["variance"])

    def choose(
        self,
        node_info: Dict[str, Dict[str, CondGaussParams]],
        pvals: List[Union[str, float]],
    ) -> float:
        """
        Return value from ConditionalLogit node
        params:
        node_info: nodes info from distributions
        pvals: parent values
        """

        cond_mean, variance = self.get_dist(node_info, pvals)
        if np.isnan(cond_mean) or np.isnan(variance):
            return np.nan

        return random.gauss(cond_mean, variance)

    def predict(
        self,
        node_info: Dict[str, Dict[str, CondGaussParams]],
        pvals: List[Union[str, float]],
    ) -> float:
        """
        Return value from ConditionalLogit node
        params:
        node_info: nodes info from distributions
        pvals: parent values
        """

        dispvals = []
        lgpvals = []
        for pval in pvals:
            if isinstance(pval, str):
                dispvals.append(pval)
            else:
                lgpvals.append(pval)

        lgdistribution = node_info["hybcprob"][str(dispvals)]

        if self.cont_parents:
            flag = False
            for el in lgpvals:
                if str(el) == "nan":
                    flag = True
                    break
            if flag:
                return np.nan
            else:
                if lgdistribution["regressor"]:
                    model = lgdistribution["regressor_obj"]
                    return model.predict(np.array(lgpvals).reshape(1, -1))[0]
                else:
                    return np.nan
        else:
            return lgdistribution["mean"]



import os
import pickle
from typing import Union

from bamt.config import config

STORAGE = config.get(
    "NODES", "models_storage", fallback="models_storage is not defined"
)


class BaseNode(object):
    """
    Base class for nodes.
    """

    def __init__(self, name: str):
        """
        :param name: name for node (taken from column name)
        type: node type
        disc_parents: list with discrete parents
        cont_parents: list with continuous parents
        children: node's children
        """
        self.name = name
        self.type = "abstract"

        self.disc_parents = []
        self.cont_parents = []
        self.children = []

    def __repr__(self):
        return f"{self.name}"

    def __eq__(self, other):
        if not isinstance(other, BaseNode):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return (
            self.name == other.name
            and self.type == other.type
            and self.disc_parents == other.disc_parents
            and self.cont_parents == other.cont_parents
            and self.children == other.children
        )

    @staticmethod
    def choose_serialization(model) -> Union[str, Exception]:
        try:
            ex_b = pickle.dumps(model, protocol=4)
            model_ser = ex_b.decode("latin1").replace("'", '"')

            if type(model).__name__ == "CatBoostRegressor":
                a = model_ser.encode("latin1")
            else:
                a = model_ser.replace('"', "'").encode("latin1")

            classifier_body = pickle.loads(a)
            return "pickle"
        except Exception as ex:
            return ex

    @staticmethod
    def get_path_joblib(node_name: str, specific: str = "") -> str:
        """
        Args:
            node_name: name of node
            specific: more specific unique name for node.
            For example, combination.

        Return:
            Path to save a joblib file.
        """
        if not isinstance(specific, str):
            specific = str(specific)

        index = str(int(os.listdir(STORAGE)[-1]))
        path_to_check = os.path.join(STORAGE, index, f"{node_name.replace(' ', '_')}")

        if not os.path.isdir(path_to_check):
            os.makedirs(os.path.join(STORAGE, index, f"{node_name.replace(' ', '_')}"))

        path = os.path.abspath(
            os.path.join(path_to_check, f"{specific}.joblib.compressed")
        )
        return path

    @staticmethod
    def get_dist(node_info, pvals):
        pass



import warnings

import numpy as np
import pandas as pd
import sklearn.preprocessing

NONE_REPLACEMENT = -32768


def information_mutual_conditional(
    x,
    y,
    z,
    cartesian_product=False,
    base=2,
    fill_value=-1,
    estimator="ML",
    alphabet_x=None,
    alphabet_y=None,
    Alphabet_Z=None,
    keep_dims=False,
):
    x, fill_value_X = _sanitise_array_input(x, fill_value)
    y, fill_value_Y = _sanitise_array_input(y, fill_value)
    z, fill_value_Z = _sanitise_array_input(z, fill_value)
    if alphabet_x is not None:
        alphabet_x, fill_value_Alphabet_X = _sanitise_array_input(
            alphabet_x, fill_value
        )
        alphabet_x, _ = _autocreate_alphabet(alphabet_x, fill_value_Alphabet_X)
    else:
        alphabet_x, fill_value_Alphabet_X = _autocreate_alphabet(x, fill_value_X)
    if alphabet_y is not None:
        alphabet_y, fill_value_Alphabet_Y = _sanitise_array_input(
            alphabet_y, fill_value
        )
        alphabet_y, _ = _autocreate_alphabet(alphabet_y, fill_value_Alphabet_Y)
    else:
        alphabet_y, fill_value_Alphabet_Y = _autocreate_alphabet(y, fill_value_Y)
    if Alphabet_Z is not None:
        Alphabet_Z, fill_value_Alphabet_Z = _sanitise_array_input(
            Alphabet_Z, fill_value
        )
        Alphabet_Z, _ = _autocreate_alphabet(Alphabet_Z, fill_value_Alphabet_Z)
    else:
        Alphabet_Z, fill_value_Alphabet_Z = _autocreate_alphabet(z, fill_value_Z)

    if x.size == 0:
        raise ValueError("arg X contains no elements")
    if y.size == 0:
        raise ValueError("arg Y contains no elements")
    if z.size == 0:
        raise ValueError("arg Z contains no elements")
    if np.any(_isnan(x)):
        raise ValueError("arg X contains NaN values")
    if np.any(_isnan(y)):
        raise ValueError("arg Y contains NaN values")
    if np.any(_isnan(z)):
        raise ValueError("arg Z contains NaN values")
    if alphabet_x.size == 0:
        raise ValueError("arg Alphabet_X contains no elements")
    if np.any(_isnan(alphabet_x)):
        raise ValueError("arg Alphabet_X contains NaN values")
    if alphabet_y.size == 0:
        raise ValueError("arg Alphabet_Y contains no elements")
    if np.any(_isnan(alphabet_y)):
        raise ValueError("arg Alphabet_Y contains NaN values")
    if Alphabet_Z.size == 0:
        raise ValueError("arg Alphabet_Z contains no elements")
    if np.any(_isnan(Alphabet_Z)):
        raise ValueError("arg Alphabet_Z contains NaN values")
    if _isnan(fill_value_X):
        raise ValueError("fill value for arg X is NaN")
    if _isnan(fill_value_Y):
        raise ValueError("fill value for arg Y is NaN")
    if _isnan(fill_value_Z):
        raise ValueError("fill value for arg Z is NaN")
    if x.shape[:-1] != alphabet_x.shape[:-1]:
        raise ValueError("leading dimensions of args X and Alphabet_X do not " "match")
    if y.shape[:-1] != alphabet_y.shape[:-1]:
        raise ValueError("leading dimensions of args Y and Alphabet_Y do not " "match")
    if z.shape[:-1] != Alphabet_Z.shape[:-1]:
        raise ValueError("leading dimensions of args Z and Alphabet_Z do not " "match")
    if not cartesian_product and (x.shape != y.shape or x.shape != z.shape):
        raise ValueError("dimensions of args X, Y, Z do not match")
    if cartesian_product and (x.shape[-1] != y.shape[-1] or x.shape[-1] != z.shape[-1]):
        raise ValueError("trailing dimensions of args X, Y, Z do not match")
    if not (np.isscalar(base) and np.isreal(base) and base > 0):
        raise ValueError("arg base not a positive real-valued scalar")

    S, fill_value = _map_observations_to_integers(
        (x, alphabet_x, y, alphabet_y, z, Alphabet_Z),
        (
            fill_value_X,
            fill_value_Alphabet_X,
            fill_value_Y,
            fill_value_Alphabet_Y,
            fill_value_Z,
            fill_value_Alphabet_Z,
        ),
    )
    x, alphabet_x, y, alphabet_y, z, Alphabet_Z = S

    if not cartesian_product:
        I = np.empty(x.shape[:-1])
        if I.ndim > 0:
            I[:] = np.NaN
        else:
            I = np.float64(np.NaN)
    else:
        shapeI_Z = z.shape[:-1]
        z = np.reshape(z, (-1, z.shape[-1]))
        Alphabet_Z = np.reshape(Alphabet_Z, (-1, Alphabet_Z.shape[-1]))
        I = []
        for i in range(z.shape[0]):

            def f(X, Y, Alphabet_X, Alphabet_Y):
                return information_mutual_conditional(
                    X,
                    Y,
                    z[i],
                    False,
                    base,
                    fill_value,
                    estimator,
                    Alphabet_X,
                    Alphabet_Y,
                    Alphabet_Z[i],
                )

            I.append(_cartesian_product_apply(x, y, f, alphabet_x, alphabet_y))
        shapeI_XY = I[0].shape
        if len(shapeI_Z) == 0:
            I = np.array(I)[0].reshape(shapeI_XY)
        else:
            I = np.array(I)
            I = np.rollaxis(I, 0, len(I.shape))
            I = I.reshape(np.append(shapeI_XY, shapeI_Z).astype("int"))
        return I

    # Re-shape H, X,Y,Z so that we may handle multi-dimensional arrays
    # equivalently and iterate across 0th axis
    x = np.reshape(x, (-1, x.shape[-1]))
    y = np.reshape(y, (-1, y.shape[-1]))
    z = np.reshape(z, (-1, z.shape[-1]))
    alphabet_x = np.reshape(alphabet_x, (-1, alphabet_x.shape[-1]))
    alphabet_y = np.reshape(alphabet_y, (-1, alphabet_y.shape[-1]))
    Alphabet_Z = np.reshape(Alphabet_Z, (-1, Alphabet_Z.shape[-1]))
    orig_shape_I = I.shape
    I = np.reshape(I, (-1, 1))

    for i in range(x.shape[0]):
        I_ = (
            entropy_joint(
                np.vstack((x[i], z[i])),
                base,
                fill_value,
                estimator,
                _vstack_pad((alphabet_x[i], Alphabet_Z[i]), fill_value),
            )
            + entropy_joint(
                np.vstack((y[i], z[i])),
                base,
                fill_value,
                estimator,
                _vstack_pad((alphabet_y[i], Alphabet_Z[i]), fill_value),
            )
            - entropy_joint(
                np.vstack((x[i], y[i], z[i])),
                base,
                fill_value,
                estimator,
                _vstack_pad((alphabet_x[i], alphabet_y[i], Alphabet_Z[i]), fill_value),
            )
            - entropy_joint(z[i], base, fill_value, estimator, Alphabet_Z[i])
        )
        I[i] = I_

    # Reverse re-shaping
    I = np.reshape(I, orig_shape_I)

    if keep_dims and not cartesian_product:
        I = I[..., np.newaxis]

    return I


def information_mutual(
    X,
    Y=None,
    cartesian_product=False,
    base=2,
    fill_value=-1,
    estimator="ML",
    Alphabet_X=None,
    Alphabet_Y=None,
    keep_dims=False,
):
    H_conditional = entropy_conditional(
        X, Y, cartesian_product, base, fill_value, estimator, Alphabet_X, Alphabet_Y
    )
    H = entropy(X, base, fill_value, estimator, Alphabet_X)

    H = np.reshape(
        H, np.append(H.shape, np.ones(H_conditional.ndim - H.ndim)).astype("int")
    )

    H = H - H_conditional

    if keep_dims and not cartesian_product:
        H = H[..., np.newaxis]

    return H


def entropy_pmf(P, base=2, require_valid_pmf=True, keep_dims=False):
    P, _ = _sanitise_array_input(P)

    if P.size == 0:
        raise ValueError("arg P contains no elements")
    if np.any(_isnan(P)):
        raise ValueError("arg P contains NaN values")
    if np.any(np.logical_or(P < 0, P > 1)):
        raise ValueError("arg P contains values outside unit interval")
    if require_valid_pmf and not np.allclose(np.sum(P, axis=-1), 1):
        raise ValueError("arg P does not sum to unity across last axis")
    if not (np.isscalar(base) and np.isreal(base) and base > 0):
        raise ValueError("arg base not a positive real-valued scalar")

    H = -np.sum(P * np.log2(P + np.spacing(0)), axis=-1)
    H = H / np.log2(base)

    if keep_dims:
        H = H[..., np.newaxis]

    return H


def entropy_joint(
    X, base=2, fill_value=-1, estimator="ML", Alphabet_X=None, keep_dims=False
):
    X, fill_value_X = _sanitise_array_input(X, fill_value)
    if Alphabet_X is not None:
        Alphabet_X, fill_value_Alphabet_X = _sanitise_array_input(
            Alphabet_X, fill_value
        )
        Alphabet_X, _ = _autocreate_alphabet(Alphabet_X, fill_value_Alphabet_X)
    else:
        Alphabet_X, fill_value_Alphabet_X = _autocreate_alphabet(X, fill_value_X)

    if X.size == 0:
        raise ValueError("arg X contains no elements")
    if np.any(_isnan(X)):
        raise ValueError("arg X contains NaN values")
    if Alphabet_X.size == 0:
        raise ValueError("arg Alphabet_X contains no elements")
    if np.any(_isnan(Alphabet_X)):
        raise ValueError("arg Alphabet_X contains NaN values")
    if _isnan(fill_value_X):
        raise ValueError("fill value for arg X is NaN")
    if X.shape[:-1] != Alphabet_X.shape[:-1]:
        raise ValueError("leading dimensions of args X and Alphabet_X do not " "match")
    if not (np.isscalar(base) and np.isreal(base) and base > 0):
        raise ValueError("arg base not a positive real-valued scalar")

    S, fill_value = _map_observations_to_integers(
        (X, Alphabet_X), (fill_value_X, fill_value_Alphabet_X)
    )
    X, Alphabet_X = S

    X = np.reshape(X, (-1, X.shape[-1]))
    Alphabet_X = np.reshape(Alphabet_X, (-1, Alphabet_X.shape[-1]))

    _verify_alphabet_sufficiently_large(X, Alphabet_X, fill_value)

    for i in range(X.shape[0]):
        X = X[:, X[i].argsort(kind="mergesort")]

    B = np.any(X[:, 1:] != X[:, :-1], axis=0)
    I = np.append(np.where(B), X.shape[1] - 1)
    L = np.diff(np.append(-1, I))

    alphabet_X = X[:, I]
    if estimator != "ML":
        n_additional_empty_bins = _determine_number_additional_empty_bins(
            L, alphabet_X, Alphabet_X, fill_value
        )
    else:
        n_additional_empty_bins = 0
    L, _ = _remove_counts_at_fill_value(L, alphabet_X, fill_value)
    if not np.any(L):
        return np.float64(np.NaN)

    # P_0 is the probability mass assigned to each additional empty bin
    P, P_0 = _estimate_probabilities(L, estimator, n_additional_empty_bins)
    H_0 = n_additional_empty_bins * P_0 * -np.log2(P_0 + np.spacing(0)) / np.log2(base)
    H = entropy_pmf(P, base, require_valid_pmf=False) + H_0

    if keep_dims:
        H = H[..., np.newaxis]

    return H


def entropy_conditional(
    X,
    Y=None,
    cartesian_product=False,
    base=2,
    fill_value=-1,
    estimator="ML",
    Alphabet_X=None,
    Alphabet_Y=None,
    keep_dims=False,
):
    if Y is None:
        Y = X
        cartesian_product = True
        Alphabet_Y = Alphabet_X

    X, fill_value_X = _sanitise_array_input(X, fill_value)
    Y, fill_value_Y = _sanitise_array_input(Y, fill_value)
    if Alphabet_X is not None:
        Alphabet_X, fill_value_Alphabet_X = _sanitise_array_input(
            Alphabet_X, fill_value
        )
        Alphabet_X, _ = _autocreate_alphabet(Alphabet_X, fill_value_Alphabet_X)
    else:
        Alphabet_X, fill_value_Alphabet_X = _autocreate_alphabet(X, fill_value_X)
    if Alphabet_Y is not None:
        Alphabet_Y, fill_value_Alphabet_Y = _sanitise_array_input(
            Alphabet_Y, fill_value
        )
        Alphabet_Y, _ = _autocreate_alphabet(Alphabet_Y, fill_value_Alphabet_Y)
    else:
        Alphabet_Y, fill_value_Alphabet_Y = _autocreate_alphabet(Y, fill_value_Y)

    if X.size == 0:
        raise ValueError("arg X contains no elements")
    if Y.size == 0:
        raise ValueError("arg Y contains no elements")
    if np.any(_isnan(X)):
        raise ValueError("arg X contains NaN values")
    if np.any(_isnan(Y)):
        raise ValueError("arg Y contains NaN values")
    if Alphabet_X.size == 0:
        raise ValueError("arg Alphabet_X contains no elements")
    if np.any(_isnan(Alphabet_X)):
        raise ValueError("arg Alphabet_X contains NaN values")
    if Alphabet_Y.size == 0:
        raise ValueError("arg Alphabet_Y contains no elements")
    if np.any(_isnan(Alphabet_Y)):
        raise ValueError("arg Alphabet_Y contains NaN values")
    if _isnan(fill_value_X):
        raise ValueError("fill value for arg X is NaN")
    if _isnan(fill_value_Y):
        raise ValueError("fill value for arg Y is NaN")
    if X.shape[:-1] != Alphabet_X.shape[:-1]:
        raise ValueError("leading dimensions of args X and Alphabet_X do not " "match")
    if Y.shape[:-1] != Alphabet_Y.shape[:-1]:
        raise ValueError("leading dimensions of args Y and Alphabet_Y do not " "match")
    if not cartesian_product and X.shape != Y.shape:
        raise ValueError("dimensions of args X and Y do not match")
    if cartesian_product and X.shape[-1] != Y.shape[-1]:
        raise ValueError("trailing dimensions of args X and Y do not match")
    if not (np.isscalar(base) and np.isreal(base) and base > 0):
        raise ValueError("arg base not a positive real-valued scalar")

    S, fill_value = _map_observations_to_integers(
        (X, Alphabet_X, Y, Alphabet_Y),
        (fill_value_X, fill_value_Alphabet_X, fill_value_Y, fill_value_Alphabet_Y),
    )
    X, Alphabet_X, Y, Alphabet_Y = S

    if not cartesian_product:
        H = np.empty(X.shape[:-1])
        if H.ndim > 0:
            H[:] = np.NaN
        else:
            H = np.float64(np.NaN)
    else:

        def f(X, Y, Alphabet_X, Alphabet_Y):
            return entropy_conditional(
                X, Y, False, base, fill_value, estimator, Alphabet_X, Alphabet_Y
            )

        return _cartesian_product_apply(X, Y, f, Alphabet_X, Alphabet_Y)

    # Re-shape H, X and Y, so that we may handle multi-dimensional arrays
    # equivalently and iterate across 0th axis
    X = np.reshape(X, (-1, X.shape[-1]))
    Y = np.reshape(Y, (-1, Y.shape[-1]))
    Alphabet_X = np.reshape(Alphabet_X, (-1, Alphabet_X.shape[-1]))
    Alphabet_Y = np.reshape(Alphabet_Y, (-1, Alphabet_Y.shape[-1]))
    orig_shape_H = H.shape
    H = np.reshape(H, (-1, 1))

    for i in range(X.shape[0]):
        H[i] = entropy_joint(
            np.vstack((X[i], Y[i])),
            base,
            fill_value,
            estimator,
            _vstack_pad((Alphabet_X[i], Alphabet_Y[i]), fill_value),
        ) - entropy(Y[i], base, fill_value, estimator, Alphabet_Y[i])

    # Reverse re-shaping
    H = np.reshape(H, orig_shape_H)

    if keep_dims and not cartesian_product:
        H = H[..., np.newaxis]

    return H


def entropy(X, base=2, fill_value=-1, estimator="ML", Alphabet_X=None, keep_dims=False):
    X, fill_value_X = _sanitise_array_input(X, fill_value)
    if Alphabet_X is not None:
        Alphabet_X, fill_value_Alphabet_X = _sanitise_array_input(
            Alphabet_X, fill_value
        )
        Alphabet_X, _ = _autocreate_alphabet(Alphabet_X, fill_value_Alphabet_X)
    else:
        Alphabet_X, fill_value_Alphabet_X = _autocreate_alphabet(X, fill_value_X)

    if X.size == 0:
        raise ValueError("arg X contains no elements")
    if np.any(_isnan(X)):
        raise ValueError("arg X contains NaN values")
    if Alphabet_X.size == 0:
        raise ValueError("arg Alphabet_X contains no elements")
    if np.any(_isnan(Alphabet_X)):
        raise ValueError("arg Alphabet_X contains NaN values")
    if _isnan(fill_value_X):
        raise ValueError("fill value for arg X is NaN")
    if X.shape[:-1] != Alphabet_X.shape[:-1]:
        raise ValueError("leading dimensions of args X and Alphabet_X do not " "match")
    if not (np.isscalar(base) and np.isreal(base) and base > 0):
        raise ValueError("arg base not a positive real-valued scalar")

    S, fill_value = _map_observations_to_integers(
        (X, Alphabet_X), (fill_value_X, fill_value_Alphabet_X)
    )
    X, Alphabet_X = S

    H = np.empty(X.shape[:-1])
    if H.ndim > 0:
        H[:] = np.NaN
    else:
        H = np.float64(np.NaN)

    # Re-shape H and X, so that we may handle multi-dimensional arrays
    # equivalently and iterate across 0th axis
    X = np.reshape(X, (-1, X.shape[-1]))
    Alphabet_X = np.reshape(Alphabet_X, (-1, Alphabet_X.shape[-1]))
    orig_shape_H = H.shape
    H = np.reshape(H, (-1, 1))

    _verify_alphabet_sufficiently_large(X, Alphabet_X, fill_value)

    # NB: This is not joint entropy. Elements in each row are sorted
    # independently
    X = np.sort(X, axis=1)

    # Compute symbol run-lengths
    # Compute symbol change indicators
    B = X[:, 1:] != X[:, :-1]
    for i in range(X.shape[0]):
        # Obtain symbol change positions
        I = np.append(np.where(B[i]), X.shape[1] - 1)
        # Compute run lengths
        L = np.diff(np.append(-1, I))

        alphabet_X = X[i, I]
        if estimator != "ML":
            n_additional_empty_bins = _determine_number_additional_empty_bins(
                L, alphabet_X, Alphabet_X[i], fill_value
            )
        else:
            n_additional_empty_bins = 0
        L, _ = _remove_counts_at_fill_value(L, alphabet_X, fill_value)
        if not np.any(L):
            continue

        # P_0 is the probability mass assigned to each additional empty bin
        P, P_0 = _estimate_probabilities(L, estimator, n_additional_empty_bins)
        H_0 = (
            n_additional_empty_bins
            * P_0
            * -np.log2(P_0 + np.spacing(0))
            / np.log2(base)
        )
        H[i] = entropy_pmf(P, base, require_valid_pmf=False) + H_0

    # Reverse re-shaping
    H = np.reshape(H, orig_shape_H)

    if keep_dims:
        H = H[..., np.newaxis]

    return H


def _autocreate_alphabet(X, fill_value):
    Lengths = np.apply_along_axis(lambda x: np.unique(x).size, axis=-1, arr=X)
    max_length = np.max(Lengths)

    def pad_with_fillvalue(x):
        return np.append(x, np.tile(fill_value, int(max_length - x.size)))

    Alphabet = np.apply_along_axis(
        lambda x: pad_with_fillvalue(np.unique(x)), axis=-1, arr=X
    )
    return (Alphabet, fill_value)


def _sanitise_array_input(x, fill_value=-1):
    # Avoid Python 3 issues with numpy arrays containing None elements
    if np.any(np.equal(x, None)) or fill_value is None:
        x = np.array(x, copy=False)
        assert np.all(x != NONE_REPLACEMENT)
        m = np.equal(x, None)
        x = np.where(m, NONE_REPLACEMENT, x)
    if fill_value is None:
        x = np.array(x, copy=False)
        fill_value = NONE_REPLACEMENT

    if isinstance(x, (pd.core.frame.DataFrame, pd.core.series.Series)):
        # Create masked array, honouring Dataframe/Series missing entries
        # NB: We transpose for convenience, so that quantities are computed for
        # each column
        x = np.ma.MaskedArray(x, x.isnull())

    if isinstance(x, np.ma.MaskedArray):
        fill_value = x.fill_value

        if np.any(x == fill_value):
            warnings.warn("Masked array contains data equal to fill value")

        if x.dtype.kind in ("S", "U"):
            kind = x.dtype.kind
            current_dtype_len = int(x.dtype.str.split(kind)[1])
            if current_dtype_len < len(fill_value):
                # Fix numpy's broken array string type behaviour which causes
                # X.filled() placeholder entries to be no longer than
                # non-placeholder entries
                warnings.warn(
                    "Changing numpy array dtype internally to "
                    "accommodate fill_value string length"
                )
                m = x.mask
                x = np.array(x.filled(), dtype=kind + str(len(fill_value)))
                x[m] = fill_value
            else:
                x = x.filled()
        else:
            x = x.filled()
    else:
        x = np.array(x, copy=False)

    if x.dtype.kind not in "biufcmMOSUV":
        raise TypeError("Unsupported array dtype")

    if x.size == 1 and x.ndim == 0:
        x = np.array((x,))

    return x, np.array(fill_value)


def _map_observations_to_integers(Symbol_matrices, Fill_values):
    assert len(Symbol_matrices) == len(Fill_values)
    FILL_VALUE = -1
    if np.any([A.dtype != "int" for A in Symbol_matrices]) or np.any(
        np.array(Fill_values) != FILL_VALUE
    ):
        L = sklearn.preprocessing.LabelEncoder()
        F = [np.atleast_1d(v) for v in Fill_values]
        L.fit(np.concatenate([A.ravel() for A in Symbol_matrices] + F))
        Symbol_matrices = [
            L.transform(A.ravel()).reshape(A.shape) for A in Symbol_matrices
        ]
        Fill_values = [L.transform(np.atleast_1d(f)) for f in Fill_values]

        for A, f in zip(Symbol_matrices, Fill_values):
            assert not np.any(A == FILL_VALUE)
            A[A == f] = FILL_VALUE

    assert np.all([A.dtype == "int" for A in Symbol_matrices])
    return Symbol_matrices, FILL_VALUE


def _isnan(X):
    X = np.array(X, copy=False)
    if X.dtype in ("int", "float"):
        return np.isnan(X)
    else:
        f = np.vectorize(_isnan_element)
        return f(X)


def _isnan_element(x):
    if isinstance(x, type(np.nan)):
        return np.isnan(x)
    else:
        return False


def _determine_number_additional_empty_bins(
    Counts, Alphabet, Full_Alphabet, fill_value
):
    alphabet_sizes = np.sum(np.atleast_2d(Full_Alphabet) != fill_value, axis=-1)
    if np.any(alphabet_sizes != fill_value):
        joint_alphabet_size = np.prod(alphabet_sizes[alphabet_sizes > 0])
        if joint_alphabet_size <= 0:
            raise ValueError(
                "Numerical overflow detected. Joint alphabet " "size too large."
            )
    else:
        joint_alphabet_size = 0
    return joint_alphabet_size - np.sum(
        np.all(np.atleast_2d(Alphabet) != fill_value, axis=0)
    )


def _estimate_probabilities(Counts, estimator, n_additional_empty_bins=0):
    assert np.sum(Counts) > 0
    assert np.all(Counts.astype("int") == Counts)
    assert n_additional_empty_bins >= 0
    Counts = Counts.astype("int")

    if isinstance(estimator, str):
        estimator = estimator.upper().replace(" ", "")

    if np.isreal(estimator) or estimator in ("ML", "PERKS", "MINIMAX"):
        if np.isreal(estimator):
            alpha = estimator
        elif estimator == "PERKS":
            alpha = 1.0 / (Counts.size + n_additional_empty_bins)
        elif estimator == "MINIMAX":
            alpha = np.sqrt(np.sum(Counts)) / (Counts.size + n_additional_empty_bins)
        else:
            alpha = 0
        Theta = (Counts + alpha) / (
            1.0 * np.sum(Counts) + alpha * (Counts.size + n_additional_empty_bins)
        )
        # Theta_0 is the probability mass assigned to each additional empty bin
        if n_additional_empty_bins > 0:
            Theta_0 = alpha / (
                1.0 * np.sum(Counts) + alpha * (Counts.size + n_additional_empty_bins)
            )
        else:
            Theta_0 = 0


def _remove_counts_at_fill_value(Counts, Alphabet, fill_value):
    I = np.any(np.atleast_2d(Alphabet) == fill_value, axis=0)
    if np.any(I):
        Counts = Counts[~I]
        Alphabet = Alphabet.T[~I].T
    return (Counts, Alphabet)


def _cartesian_product_apply(X, Y, function, Alphabet_X=None, Alphabet_Y=None):
    assert X.ndim > 0 and Y.ndim > 0
    assert X.size > 0 and Y.size > 0
    if Alphabet_X is not None or Alphabet_Y is not None:
        assert Alphabet_X.ndim > 0 and Alphabet_Y.ndim > 0
        assert Alphabet_X.size > 0 and Alphabet_Y.size > 0

    H = np.empty(np.append(X.shape[:-1], Y.shape[:-1]).astype("int"))
    if H.ndim > 0:
        H[:] = np.NaN
    else:
        H = np.float64(np.NaN)

    X = np.reshape(X, (-1, X.shape[-1]))
    Y = np.reshape(Y, (-1, Y.shape[-1]))
    if Alphabet_X is not None or Alphabet_Y is not None:
        Alphabet_X = np.reshape(Alphabet_X, (-1, Alphabet_X.shape[-1]))
        Alphabet_Y = np.reshape(Alphabet_Y, (-1, Alphabet_Y.shape[-1]))
    orig_shape_H = H.shape
    H = np.reshape(H, (-1, 1))

    n = 0
    for i in range(X.shape[0]):
        for j in range(Y.shape[0]):
            if Alphabet_X is not None or Alphabet_Y is not None:
                H[n] = function(X[i], Y[j], Alphabet_X[i], Alphabet_Y[j])
            else:
                H[n] = function(X[i], Y[j])
            n = n + 1

    # Reverse re-shaping
    H = np.reshape(H, orig_shape_H)

    return H


def _verify_alphabet_sufficiently_large(X, Alphabet, fill_value):
    assert not np.any(X == np.array(None))
    assert not np.any(Alphabet == np.array(None))
    for i in range(X.shape[0]):
        I = X[i] != fill_value
        J = Alphabet[i] != fill_value
        # NB: This causes issues when both arguments contain None. But it is
        # always called after observations have all been mapped to integers.
        if np.setdiff1d(X[i, I], Alphabet[i, J]).size > 0:
            raise ValueError(
                "provided alphabet does not contain all observed " "outcomes"
            )


def _vstack_pad(arrays, fill_value):
    max_length = max([A.shape[-1] for A in arrays])
    arrays = [
        np.append(
            A,
            np.tile(
                fill_value,
                np.append(A.shape[:-1], max_length - A.shape[-1]).astype(int),
            ),
        )
        for A in arrays
    ]
    return np.vstack(arrays)



from collections import defaultdict


def nodes_from_edges(edges: list):
    """
    Retrieves all nodes from the list of edges.
            Arguments
    ----------
    *edges* : list

    Returns
    -------
    *nodes* : list

    Effects
    -------
    None
    """
    return set().union(edges)


def edges_to_dict(edges: list):
    """
    Transfers the list of edges to the dictionary of parents.
            Arguments
    ----------
    *edges* : list

    Returns
    -------
    *parents_dict* : dict

    Effects
    -------
    None
    """
    nodes = nodes_from_edges(edges)
    parents_dict = defaultdict(list, {node: [] for node in nodes})
    parents_dict.update(
        {child: parents_dict[child] + [parent] for parent, child in edges}
    )
    return parents_dict



from .base import BaseNetwork


class ContinuousBN(BaseNetwork):
    """
    Bayesian Network with Continuous Types of Nodes
    """

    def __init__(self, use_mixture: bool = False):
        super(ContinuousBN, self).__init__()
        self.type = "Continuous"
        self._allowed_dtypes = ["cont"]
        self.has_logit = None
        self.use_mixture = use_mixture
        self.scoring_function = ""
